// Database.cpp: implementation of the Database class.
//
//////////////////////////////////////////////////////////////////////

#include "Database.h"
#include <stdio.h>

Database::Database()
{
	strcpy(m_pszServerHost, "127.0.0.1");
	m_nServerPort = 0;
	strcpy(m_pszID, "MPICH");
}

Database::~Database()
{
}

void Database::SetID(char *pszID)
{
	strcpy(m_pszID, pszID);
}

void Database::Init()
{
	int len;
	char pszTemp[100];
	bool bFound = false;

	if (GetEnvironmentVariable("MPICH_DBS", pszTemp, 100)
	{
		char *token;
		token = strtok(pszTemp, ":");
		if (token != NULL)
			strcpy(m_pszServerHost, token);
		token = strtok(NULL, " \n");
		if (token != NULL)
			m_nServerPort = atoi(token);
		bFound = true;
	}
	else
	{
		if (GetEnvironmentVariable("MPICH_DBS_HOST", m_pszServerHost, 100) &&
			GetEnvironmentVariable("MPICH_DBS_PORT", pszTemp, 100))
		{
			m_nServerPort = atoi(pszTemp);
			bFound = true;
		}
	}

	if (bFound)
	{
		int ret_val;
		char ack, cmd;
		WSAEVENT sock_event;
		SOCKET sock;
		
		m_nServerPort = atoi(pszTemp);

		// create the event
		sock_event = WSACreateEvent();
		if (sock_event == WSA_INVALID_EVENT)
			dbs_error("WSACreateEvent failed in Database::Init()", WSAGetLastError());

		// create the socket
		sock = socket(PF_INET, SOCK_STREAM, 0);
		if (sock == INVALID_SOCKET)
			dbs_error("socket failed in Database::Init()", WSAGetLastError());
		
		// connect to server
		ret_val = NT_Tcp_connect(sock, m_pszServerHost, m_nServerPort);
		if (ret_val)
			dbs_error("Database::Init: NT_Tcp_connect failed", ret_val);
		
		if (WSAEventSelect(sock, sock_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
			dbs_error("Database::Init: WSAEventSelect failed", WSAGetLastError());

		// send id
		len = strlen(m_pszID)+1;
		if (SendBlocking(sock, (char*)&len, sizeof(int), 0) == SOCKET_ERROR)
			dbs_error("Database::Init: send len failed", WSAGetLastError());
		if (SendBlocking(sock, m_pszID, len, 0) == SOCKET_ERROR)
			dbs_error("Database::Init: send pszID failed", WSAGetLastError());

		// send command
		cmd = MPI_DBS_CMD_EXISTS;
		if (SendBlocking(sock, &cmd, 1, 0) == SOCKET_ERROR)
			dbs_error("Database::Init: send cmd failed", WSAGetLastError());
		
		// receive response
		ret_val = ReceiveBlocking(sock, sock_event, &ack, 1, 0);
		if (ret_val)
			dbs_error("Database::Init: recv ack failed", WSAGetLastError());

		if (ack != MPI_DBS_SUCCESS)
			dbs_error("Unable to contact mpi database server\n", 1);

		// close socket
		NT_Tcp_closesocket(sock, sock_event);
	}
}

Database& Database::operator =(Database &db)
{
	if (this != &db)
	{
		m_nServerPort = db.m_nServerPort;
		strcpy(m_pszID, db.m_pszID);
		strcpy(m_pszServerHost, db.m_pszServerHost);
	}
	return *this;
}

int Database::Delete()
{
	int ret_val, len;
	char cmd, ack;
	WSAEVENT sock_event;
	SOCKET sock;
	
	// create the event
	sock_event = WSACreateEvent();
	if (sock_event == WSA_INVALID_EVENT)
		dbs_error("WSACreateEvent failed in Database::Get()", WSAGetLastError());

	// create the socket
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock == INVALID_SOCKET)
		dbs_error("socket failed in Database::Get()", WSAGetLastError());

	// connect to server
	ret_val = NT_Tcp_connect(sock, m_pszServerHost, m_nServerPort);
	if (ret_val)
		dbs_error("Database::Delete: NT_Tcp_connect failed", ret_val);
	
	if (WSAEventSelect(sock, sock_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
		dbs_error("Database::Delete: WSAEventSelect failed", WSAGetLastError());
	
	// send id
	len = strlen(m_pszID)+1;
	if (SendBlocking(sock, (char*)&len, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Delete: send len failed", WSAGetLastError());
	if (SendBlocking(sock, m_pszID, len, 0) == SOCKET_ERROR)
		dbs_error("Database::Delete: send pszID failed", WSAGetLastError());

	// send delete command
	cmd = MPI_DBS_CMD_DELETE;
	if (SendBlocking(sock, &cmd, 1, 0) == SOCKET_ERROR)
		dbs_error("Database::Delete: send cmd failed", WSAGetLastError());
	
	// receive ack
	ret_val = ReceiveBlocking(sock, sock_event, &ack, 1, 0);
	if (ret_val)
		dbs_error("Database::Delete: recv ack failed", WSAGetLastError());

	// close socket
	NT_Tcp_closesocket(sock, sock_event);

	return ack == MPI_DBS_SUCCESS ? MPI_DBS_SUCCESS : MPI_DBS_FAIL;
}

int Database::Get(char *pszKey, void *pValue, int *length)
{
	int ret_val, len;
	char cmd;
	WSAEVENT sock_event;
	SOCKET sock;
	
	// create the event
	sock_event = WSACreateEvent();
	if (sock_event == WSA_INVALID_EVENT)
		dbs_error("WSACreateEvent failed in Database::Get()", WSAGetLastError());

	// create the socket
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock == INVALID_SOCKET)
		dbs_error("socket failed in Database::Get()", WSAGetLastError());

	// connect to server
	ret_val = NT_Tcp_connect(sock, m_pszServerHost, m_nServerPort);
	if (ret_val)
		dbs_error("Database::Get: NT_Tcp_connect failed", ret_val);
	
	if (WSAEventSelect(sock, sock_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
		dbs_error("Database::Get: WSAEventSelect failed", WSAGetLastError());
	
	// send id
	len = strlen(m_pszID)+1;
	if (SendBlocking(sock, (char*)&len, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Get: send len failed", WSAGetLastError());
	if (SendBlocking(sock, m_pszID, len, 0) == SOCKET_ERROR)
		dbs_error("Database::Get: send pszID failed", WSAGetLastError());

	// send get command
	cmd = MPI_DBS_CMD_GET;
	if (SendBlocking(sock, &cmd, 1, 0) == SOCKET_ERROR)
		dbs_error("Database::Get: send cmd failed", WSAGetLastError());
	
	// send key length, key, buffer length
	len = strlen(pszKey)+1;
	if (SendBlocking(sock, (char*)&len, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Get: send len failed", WSAGetLastError());
	if (SendBlocking(sock, pszKey, len, 0) == SOCKET_ERROR)
		dbs_error("Database::Get: send pszKey failed", WSAGetLastError());
	if (SendBlocking(sock, (char*)length, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Get: send length failed", WSAGetLastError());

	// receive length and buffer
	// the buffer is not sent if there isn't enough room
	ret_val = ReceiveBlocking(sock, sock_event, (char*)&len, sizeof(int), 0);
	if (ret_val)
		dbs_error("Database::Get: recv len failed", WSAGetLastError());
	if (len <= *length)
	{
		ret_val = ReceiveBlocking(sock, sock_event, (char*)pValue, len, 0);
		if (ret_val)
			dbs_error("Database::Get: recv pValue failed", WSAGetLastError());
		*length = len;
		NT_Tcp_closesocket(sock, sock_event);
		return MPI_DBS_SUCCESS;
	}
	*length = len;
	return MPI_DBS_FAIL;
}

int Database::Put(char *pszKey, void *pValue, int length, bool bPersistent)
{
	int ret_val, len;
	char cmd, ack;
	WSAEVENT sock_event;
	SOCKET sock;
	
	// create the event
	sock_event = WSACreateEvent();
	if (sock_event == WSA_INVALID_EVENT)
		dbs_error("WSACreateEvent failed in Database::Put()", WSAGetLastError());

	// create the socket
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock == INVALID_SOCKET)
		dbs_error("socket failed in Database::Put()", WSAGetLastError());

	// connect to server
	ret_val = NT_Tcp_connect(sock, m_pszServerHost, m_nServerPort);
	if (ret_val)
		dbs_error("Database::Put: NT_Tcp_connect failed", ret_val);
	
	if (WSAEventSelect(sock, sock_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
		dbs_error("Database::Put: WSAEventSelect failed", WSAGetLastError());
	
	// send id
	len = strlen(m_pszID)+1;
	if (SendBlocking(sock, (char*)&len, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send len failed", WSAGetLastError());
	if (SendBlocking(sock, m_pszID, len, 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send pszID failed", WSAGetLastError());

	// send put command
	cmd = bPersistent ? MPI_DBS_CMD_PUT_PERSISTENT : MPI_DBS_CMD_PUT_CONSUMABLE;
	if (SendBlocking(sock, &cmd, 1, 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send cmd failed", WSAGetLastError());
	
	// send key length, key, buffer length, buffer
	len = strlen(pszKey)+1;
	if (SendBlocking(sock, (char*)&len, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send len failed", WSAGetLastError());
	if (SendBlocking(sock, pszKey, len, 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send pszKey failed", WSAGetLastError());
	if (SendBlocking(sock, (char*)&length, sizeof(int), 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send length failed", WSAGetLastError());
	if (SendBlocking(sock, (char*)pValue, length, 0) == SOCKET_ERROR)
		dbs_error("Database::Put: send pValue failed", WSAGetLastError());

	// receive ack
	ret_val = ReceiveBlocking(sock, sock_event, &ack, 1, 0);
	if (ret_val)
		dbs_error("Database::Put: recv ack failed", WSAGetLastError());

	// close socket
	NT_Tcp_closesocket(sock, sock_event);

	return ack == MPI_DBS_SUCCESS ? MPI_DBS_SUCCESS : MPI_DBS_FAIL;
}

void dbs_error(char *string, int value)
{
	printf("Error %d\n   %s\n", value, string);fflush(stdout);
    
	WSACleanup();
	ExitProcess(1);
}

void dbs_error(char *string, int value, SOCKET sock, WSAEVENT sock_event)
{
	NT_Tcp_closesocket(sock, sock_event);
	printf("Error %d\n   %s\n", value, string);fflush(stdout);
    
	WSACleanup();
	ExitProcess(1);
}
